;*******************************************************************************
;
;	Filename:		ToiletSentinel.asm
;	Date:				14 Nov 2015	
;	File Version:	.1
;	Author:			Larry Cicchinelli
;	Description:
;		This program will monitor the water level in a toilet tank.
;
;		PIC I/O usage:
;			pin 2: RA5, buzzer, uses PWM1, freq = LFINTOSC/16
;			pin 3: RA4, input from tank sensor
;			pin 4: RA3, not used, ~MCLR
;			pin 5: RA2, pseudo RS232 port - output only - 300baud
;			pin 6: RA1, not used, ICSPCLK
;			pin 7: RA0, not used, ISCPDAT
;
;Note: all functions return the BSR value to PORTA
;
; A battery test is executed every time the toilet is flushed.  The test
; measures the FVR using VDD as the reference.  If FVR (1.024V) becomes greater
; than 1/2 VDD, that means the battery is low.  When that occurs, the system
; signal low battery via the buzzer.
;
; Startup
;	1) initialization of all systems in the PIC
;	2) delay 2 seconds
;	3) sound 'R' for Running
;	4) delay 2 seconds
;	5) force calibration cycle
;	6) drop through to:
; Normal operation
;	1) check for flush
;	2) goto 1)
;
; Calibration cycle
;	This occurs when the unit is first powered.
;	1) sound 'F'
;	2) Wait for the sensor to indicate that a flush is in progress (RA4=high).
;	3) Check the battery
;	4) Wait for the sensor to indicate that the water has filled the tank(RA4=low).
;	5) The elapsed time, in seconds, between 3 and 4 is sounded via the buzzer.
;			The value is then multiplied by 1.25 and stored as the reference flush
;			time.
;
; Normal flush cycle
;	1) Monitor the sensor until it indicates that a flush is in progress.
;	2) battery check
;	3) Monitor the sensor until it indicates a full tank or the time has exceeded
;		the reference flush time.
;		a) if full tank is detected:
;				sound: short, short
;				go to 1
;		b) if elapsed time exceeds reference flush time:
;				sound problem code
;				goto 1
;	NOTE: If the tank does not fill the program will continuously execute
;		steps 1 - 3.
;
; Buzzer codes:
;	calibration signals:
;		short, short, long, short	= F: start calibration flush cycle
;		battery check
;		sound the number of seconds in binary: long = 1, short = 0
;		if flush time too long		= L: short, long, short, short
;	normal operation signals
;		program Running				= R: running
;		battery check
;		short, long, long, short	= P: problem
;	battery check:
;		short								= E: battery OK
;		long, short, short, short	= B: battery low
;
;*******************************************************************************
; Processor Inclusion
		#include p12LF1572.inc

;*******************************************************************************

; PIC12LF1572 Configuration Bit Settings

; CONFIG1
; __config 0xF9E4
 __CONFIG _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_ON & _CP_OFF & _BOREN_OFF & _CLKOUTEN_OFF
; CONFIG2
; __config 0xFEFF
 __CONFIG _CONFIG2, _WRT_OFF & _PLLEN_OFF & _STVREN_ON & _BORV_LO & _LPBOREN_OFF & _LVP_ON

; misc hardware definitions
SERIAL_OUT		equ	RA2
SENSOR			equ	RA4
BUZZER			equ	RA5
BUZZ_SHORT		equ	.10				; 10ms units
BUZZ_LONG		equ	3*BUZZ_SHORT

; RAM allocations

; Common RAM
Access_RAM		udata_shr	0x70		; 16 bytes starting at address 0x70
CalibrateSecs	res	1					; nbr of seconds in calibrate flush
DebounceTime	res	1					; used in Chk.Cal
FlushSeconds	res	1					; nbr of seconds in current flush cycle
DelayCount		res	1					; used by Delay.10xMsec
BuzzTimer		res	1					; used by the buzzer routines
ShiftCounter	res	1
SerialByte		res	1					; byte being transmitted via serial.out

;*******************************************************************************
; Vectors
;*******************************************************************************

	org	0
	goto	Main

	org	0x04								; interrupt vector
	clrf	INTCON							; force all interrupts off
	retfie

;*******************************************************************************
; MAIN PROGRAM
;*******************************************************************************

MAIN_PROG CODE                      ; let linker place main program

Main
	call		init							; initialize the system hardware
	movlw		.200
	call		Delay.10xWmsec				; wait 2 seconds
	call		Buzz.Short					; tell
	call		Buzz.Long					;		program 
	call		Buzz.Short					;			Running
	movlw		.200
	call		Delay.10xWmsec				; wait 2 seconds
	call		Calibrate					; force calibrate

Check.Flush
; determine whether or not the tank is full
	BANKSEL	PORTA
	btfss		PORTA, SENSOR				; skip next if flush detected
	goto		Check.Flush					; flush not detected
; here if flush detected
	call		Check.Battery
	movlw		"E"
	call		Serial.Out
	movf		CalibrateSecs, w			; get nbr of seconds of calibrate flush
	movwf		FlushSeconds				; save for loop
; wait for full tank
Check.Flush.10
	call		Delay.1sec					; wait 1 second
	btfss		PORTA, SENSOR				; skip next if still flushing
	goto		Check.Flush.20				; goto if full tank detected
	; here if not done
	decfsz	FlushSeconds, F			; update timeout - error if timed out
	goto		Check.Flush.10				; goto if not done
; here if too long
	call		Buzz.Short					; show Problem
	call		Buzz.Long
	call		Buzz.Long
	call		Buzz.Short
	movlw		"P"
	call		Serial.Out
	call		Delay.1sec					; wait 1 second
	goto		Check.Flush
Check.Flush.20
	call		Buzz.Short					; tell full tank detected and return
	call		Buzz.Short
	movlw		"I"
	call		Serial.Out
	goto		Check.Flush
; Check.Flush


Calibrate
; Calibrate the tank
; exit with:
;		CalibrateSecs = nbr of seconds for flush cycle
Calibrate
; tell user to flush
	call		Buzz.Short
	call		Buzz.Short
	call		Buzz.Long
	call		Buzz.Short
; wait for tank to start emptying
Calibrate.40
	movlw		.30							; set up for .3sec debounce
	movwf		DebounceTime
Calibrate.45
	btfss		PORTA, SENSOR				; skip next if flush detected
	goto		Calibrate.40
	call		Delay.10ms					; delay 10ms
	decfsz	DebounceTime, F			; debounce timeout?
	goto		Calibrate.45				; no
; here when flush detected
	call		Check.Battery
	clrf		CalibrateSecs				; init nbr of seconds to flush
Calibrate.50
	call		Delay.1sec					; wait 1 second
	incf		CalibrateSecs, F			; update flush time
; test if flush time is too long
	movf		CalibrateSecs, w			; copy current time
	andlw		0xC0							; keep only upper 2 bits (.192)
	xorlw		0xC0							; W = 0 if upper 2 bits were all 1
	btfsc		STATUS, Z
	goto		Calibrate.100				; goto if flush time too long
;
	btfsc		PORTA, SENSOR				; skip next if full tank detected
	goto		Calibrate.50				; goto if not full yet
	call		Delay.1sec					; wait 1 second
	call		Buzz.Seconds				; tell flush time
; add flush time/4 to value as pad
	movf		CalibrateSecs, w			; retrieve calibrate flush time
	asrf		WREG, f						; /2
	asrf		WREG, f						; /4
	addwf		CalibrateSecs, f			; CalibrateSecs = 1.25 * CalibrateSecs
Calibrate.60
	nop						; for debugging
	return
;
Calibrate.100
	call		Buzz.Short
	call		Buzz.Long
	call		Buzz.Short
	call		Buzz.Short
	goto		Calibrate.60
; Calibrate


;===============================================================================
;									Check.Battery
;===============================================================================
Check.Battery	; ensure that the battery is still >2V
; enable the FVR - already set to 1.024V
; sound the battery low signal if the battery is < 2.048V
; IMPORTANT: When the battery drops below 2.048V, the MSB of the reading is
; equivalent to 1.048V which is the FVR value.
	BANKSEL	FVRCON
	bsf		FVRCON, FVREN
Check.Battery.5
	btfss		FVRCON, FVRRDY				; wait for
	goto		Check.Battery.5			;		FVR ready
; start a reading
	BANKSEL	ADCON0
	bsf		ADCON0, ADON				; turn A/D on
	nop										; a little extra acquiition time
	bsf		ADCON0, GO					; start a reading
; wait for A/D done
Check.Battery.10
	btfsc		ADCON0, GO					; skip next if done
	goto		Check.Battery.10			; not done
;
	btfsc		ADRESH, 7					; battery OK if bit 7 = 0
	call		Buzz.Battery.Low			; tell user low battery
; turn off the A/D and FVR to conserve power
	call		Buzz.Short					; single short is battery OK
	BANKSEL	ADCON0
	bcf		ADCON0, ADON				; turn A/D off
	BANKSEL	FVRCON
	bcf		FVRCON, FVREN				; disable FVR
	BANKSEL	PORTA							; reset BSR
	return
; Check.Battery

; last Buzz.Short is implemented by caller
Buzz.Battery.Low							; short, long, short, short
	call		Buzz.Long					; tell low Battery
	call		Buzz.Short
	goto		Buzz.Short					; this will return to the caller
; Buzz.Battery.Low


;===============================================================================
;									Initialize Hardware
;===============================================================================
init	; initialize the system hardware

; main oscillator
	BANKSEL	OSCCON						; PIEn, OSCCON, ADCONn, TRISA
	movlw		(0<<SPLLEN)|(0<<IRCF0)|(b'10'<<SCS0)
	movwf		OSCCON						; PLL disabled, 31KHz LF osc, use Internal osc

; A/D convertor
	movlw		(b'11111'<<CHS0)
	movwf		ADCON0						; select FVR
	movlw		(0<<ADFM)|(b'111'<<ADCS0)|(0<<ADPREF0)
	movwf		ADCON1						; left justify data, clk=FRC, VRPOS=VDD
	clrf		ADCON2
; note: PIEn registers are not used

; Port A
	movlw		b'00011011'
	movwf		TRISA							; RA<4,3,1,0> = input,  RA<5,2> = output
	BANKSEL	ANSELA
	clrf		ANSELA						; port A pins = digital I/O
	BANKSEL	PORTA
	clrf		PORTA
	bsf		PORTA, SERIAL_OUT			; serial port = mark = high 

; FVR
	BANKSEL	FVRCON
	movlw		1								; ADC FVR Buffer Gain is 1x,
	movwf		FVRCON						;		VADFVR = 1x VFVR = 1.024V

; PWM	set for 2kHz (LFINTOSC/16), 50%
	BANKSEL	APFCON
	movlw		1								; RA5 = PWM1
	movwf		APFCON
	BANKSEL	PWM1CLKCON
	movlw		b'10'							; select LFINTOSC, no prescaler
	movwf		PWM1CLKCON
	movlw		(0<<EN)|(1<<OE)|(0<<POL)|(b'00'<<MODE0) ; disable PWM1, output on
	movwf		PWM1CON						; active high, standard mode
	clrf		PWM1PHH						; start period at phase = 0
	clrf		PWM1PHL
	clrf		PWM1DCH						; end duty cycle at count = 8
	movlw		.8
	movwf		PWM1DCL
	clrf		PWM1PRH						; end period at count = 16
	movlw		.15							; (divides by N+1)
	movwf		PWM1PRL

	BANKSEL	PORTA
	return
; init


;===============================================================================
;									Delay Functions
;===============================================================================
; NOTE: the delay functions require a CPU clock of 31kHz == 32us
;	and since each instruction takes four clock cycles, an instruction takes
;	about 4*32usec = 128usec
;
; This function is a programmed delay of about 10ms requiring 78 instructions
; uses W
Delay.10ms
;	call										; 2
	movlw		.24							; 1
Delay.10ms.10
	decfsz	WREG,f						; mostly 1 cycle, last iteration = 2 cycles
	goto		Delay.10ms.10				; 2 cycles
	nop										; 1
	nop										; 1
	return									; 2
;	Total time = (2 + 1 + [(N-1)*3]+2 + 1 + 1 + 2)*128 = [(N-1)*3] + 9 = 9984us
; Delay.10ms


Delay.1sec
; delay 1 second
	movlw		.100
; drop thru to Delay.10xWmsec

; Delay W * 10ms
; enter with W = nbr of ms/10
; uses W, DelayCount
Delay.10xWmsec
	movwf		DelayCount					; get count for loop
Delay.10xWmsec.10
	call		Delay.10ms					; wait 10msec
	decfsz	DelayCount,f				; skip next if done
	goto		Delay.10xWmsec.10
	return
; Delay.10xWmsec


;===============================================================================
;									Buzzer Functions
;===============================================================================
; Activate the buzzer
; enter with W = nbr of ms/10
Sound.Buzzer
	BANKSEL	PWM1CLKCON
	bsf		PWM1CON, EN					; enable PWM1
	call		Delay.10xWmsec
	bcf		PWM1CON, EN					; disable PWM1
	BANKSEL	PORTA							; reset BSR
	return
; Sound.Buzzer


Buzz.Short
	movlw		BUZZ_SHORT					; get short buzzer value
Buzz.Short.10
	call		Sound.Buzzer				; turn on buzzer for short buzzer time
	movlw		BUZZ_SHORT
	goto		Delay.10xWmsec				; wait short buzzer time and return
; Buzz.Short


Buzz.Long	
	movlw		BUZZ_LONG					; get long buzzer value
	goto		Buzz.Short.10
; Buzz.Long

Buzz.Seconds
; tell user calibrate completed by sounding the flush seconds count
	; find MS 1
	movlw		.8								; 8 bits
	movwf		ShiftCounter				; save for loop
	movf		CalibrateSecs, w			; get nbr of seconds
	movwf		FlushSeconds				; copy for loop
; search for Most Significant '1'
Buzz.Seconds.10
	rlf		FlushSeconds, f
	decf		ShiftCounter, F			; update shift counter
	btfss		STATUS, C
	goto		Buzz.Seconds.10			; try next bit
; here when MS '1' bit found
	call		Buzz.Long
Buzz.Seconds.20
	movlw		BUZZ_LONG
	call		Delay.10xWmsec				; wait long buzzer time between bits
	movf		ShiftCounter, f			; done?
	btfsc		STATUS, Z					; skip if no
	return
	rlf		FlushSeconds, f
	decf		ShiftCounter, F			; update shift counter
	btfss		STATUS, C
	goto		Buzz.Seconds.30			; must be a '0'
	call		Buzz.Long
	goto		Buzz.Seconds.20
Buzz.Seconds.30
	call		Buzz.Short
	goto		Buzz.Seconds.20
; Buzz.Seconds

Serial.Out
; send a byte out of a bit-banged serial port
; enter with W = byte value to transmit - MSB first
; 9 instructions/bit = 31000/4/9 = 861baud
	movwf		SerialByte				; save the byte
	movlw		8							; 8 bits
	movwf		ShiftCounter			; save nbr of bits
	bcf		PORTA, SERIAL_OUT		; start bit
	call		Serial.Out.40			; 4 instruction cycles
	nop
Serial.Out.10
	call		Serial.Out.Delay
	rlf		SerialByte, f			; put bit into C
	btfsc		STATUS, C				; skip next if C=0
	goto		Serial.Out.20
; here if bit = 0
	bcf		PORTA, SERIAL_OUT		; send 0
	goto		Serial.Out.30
; here if bit = 1
Serial.Out.20
	bsf		PORTA, SERIAL_OUT		; send 1
	nop									; pad to make
	nop									;		same time as bit = 0
	nop
Serial.Out.30
	decfsz	ShiftCounter, f		; update bit count, skip next if done
	goto		Serial.Out.10			; do next bit
; handle stop bit
	call		Serial.Out.Delay		; make last bit
	call		Serial.Out.40			;		correct width
	nop
	bsf		PORTA, SERIAL_OUT		; stop bit is a mark
	call		Delay.10ms
Serial.Out.40
	return
	
Serial.Out.Delay						; delay to get 300 baud
	movlw		4
Serial.Out.Delay.10
	decfsz	WREG, f
	goto		Serial.Out.Delay.10
	nop
	return
; Serial.Out

	END
